/*
 * Copyright (c) 2022 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/clock_control/nrf_clock_control.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/irq.h>
#include <zephyr/logging/log.h>
#include <nrf.h>
#include <esb.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <zephyr/sys/printk.h>
#include <zephyr/ipc/ipc_service.h>


#define STACKSIZE	(1024)

LOG_MODULE_REGISTER(esb_ptx, CONFIG_ESB_PTX_APP_LOG_LEVEL);

static const struct device *ipc0_instance = DEVICE_DT_GET(DT_NODELABEL(ipc0));
static struct ipc_ept ep;

#define LED0_NODE DT_ALIAS(led0)

static const struct gpio_dt_spec led0 = GPIO_DT_SPEC_GET(LED0_NODE, gpios);

int ret;

#define NETWORK_NODES 3       // Number of nodes present in the network
#define NETWORK_PIPE 0

// transmission time for each node in us
#define TIME_TX_SINGLE_NODE_US 800
// delay added in nodes to allow the station to be ready to receive messages
#define TIME_STATION_SWITCH_RX 50
// maximum latency of the sync message
#define TIME_SYNC_LATENCY 50 
// time needed for the node to switch to radio rx mode
#define TIME_NODE_SWITCH_RX 400 
// time needed by a node to read the data and perform haptic commands
#define TIME_IFEEL_NODE_TASKS_US 10000
// extra time needed to allow all of the nodes to execute their tasks
#define TIME_EXTRA TIME_IFEEL_NODE_TASKS_US -(NETWORK_NODES/2)*TIME_TX_SINGLE_NODE_US 

// sync message period
const uint32_t TIME_SYNC = TIME_SYNC_LATENCY // max latency of sync messages
                           // same delays as in the nodes
                           + TIME_STATION_SWITCH_RX
                           #if TIME_EXTRA>0
                           + TIME_EXTRA
                           #endif
                           + (TIME_TX_SINGLE_NODE_US * NETWORK_NODES)
                           + TIME_NODE_SWITCH_RX; // wait for nodes to be ready to receive sync message



static bool ready = true;
static uint8_t n_sync = 0;
struct esb_config config = ESB_DEFAULT_CONFIG;
static struct esb_payload rx_payload;
static struct esb_payload tx_payload = ESB_CREATE_PAYLOAD(NETWORK_PIPE,	0xAA);

#define MAX_PAYLOAD_LEN 40

#define _RADIO_SHORTS_COMMON                                                   \
	(RADIO_SHORTS_READY_START_Msk | RADIO_SHORTS_END_DISABLE_Msk |         \
	 RADIO_SHORTS_ADDRESS_RSSISTART_Msk |                                  \
	 RADIO_SHORTS_DISABLED_RSSISTOP_Msk)

struct k_timer time_sync_timer;

static volatile bool radio_tx_busy = false;

static K_SEM_DEFINE(bound_sem, 0, 1);

static void ep_bound(void *priv)
{
	k_sem_give(&bound_sem);
}

static struct ipc_ept_cfg ep_cfg = {
	.name = "ep0",
	.cb = {
		.bound    = ep_bound,
		.received = NULL,
	},
};

void time_sync_expiry(struct k_timer *timer_id){
	if(config.mode != ESB_MODE_PTX)
	{
		esb_stop_rx();
		esb_disable();
		config.mode = ESB_MODE_PTX;
		esb_init(&config);
	}
	esb_start_tx();

	tx_payload.data[0] = 1;

	if(esb_write_payload(&tx_payload) == 0)
	{  
		radio_tx_busy = true;
	}

    return;

}


void my_esb_event_handler(struct esb_evt const *event)
{
	ready = true;

	switch (event->evt_id) {
	case ESB_EVENT_TX_SUCCESS:
		LOG_DBG("TX SUCCESS EVENT");
			radio_tx_busy = false;
		
		n_sync++;
		if(n_sync==100){
			gpio_pin_toggle_dt(&led0);
			n_sync = 0;
		}

		esb_disable();
		config.mode = ESB_MODE_PRX;
		esb_init(&config);
		esb_start_rx();


		break;
	case ESB_EVENT_TX_FAILED:
		LOG_DBG("TX FAILED EVENT");
		radio_tx_busy = false;
		
		n_sync++;
		if(n_sync==30){
			gpio_pin_toggle_dt(&led0);
			n_sync = 0;
		}

		break;
	case ESB_EVENT_RX_RECEIVED:
		while (esb_read_rx_payload(&rx_payload) == 0) {
			 if (rx_payload.length > 0)
                {
					// I'm not fast enough to manipulate the incoming payload,
					// so I'll send  it to the AppCore.

					// Given that it will work only with my actual ESB network nodes, I'll send some mock data in the main while loop.

					//ret = ipc_service_send(&ep, rx_payload.data, MAX_PAYLOAD_LEN);
					//printk("IPC SEND RET MSG: %d\n",ret);
				}
			printk("ID: %d\n\n", rx_payload.data[0]);
		}
		break;	
	}
}

int clocks_start(void)
{
	int err;
	int res;
	struct onoff_manager *clk_mgr;
	struct onoff_client clk_cli;

	clk_mgr = z_nrf_clock_control_get_onoff(CLOCK_CONTROL_NRF_SUBSYS_HF);
	if (!clk_mgr) {
		LOG_ERR("Unable to get the Clock manager");
		return -ENXIO;
	}

	sys_notify_init_spinwait(&clk_cli.notify);

	err = onoff_request(clk_mgr, &clk_cli);
	if (err < 0) {
		LOG_ERR("Clock request failed: %d", err);
		return err;
	}

	do {
		err = sys_notify_fetch_result(&clk_cli.notify, &res);
		if (!err && res) {
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err);

	LOG_DBG("HF clock started");
	return 0;
}

int esb_initialize(void)
{
	int err;
	/* These are arbitrary default addresses. In end user products
	 * different addresses should be used for each set of devices.
	 */
	uint8_t base_addr_0[4] = {0xE7, 0xE7, 0xE7, 0xE7};
	uint8_t base_addr_1[4] = {0xC2, 0xC2, 0xC2, 0xC2};
	uint8_t addr_prefix[8] = {0xE7, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8};

	config.payload_length = MAX_PAYLOAD_LEN;	
	config.protocol = ESB_PROTOCOL_ESB_DPL;
	config.retransmit_delay = 250;
	config.bitrate = ESB_BITRATE_2MBPS;
	config.event_handler = my_esb_event_handler;
	config.mode = ESB_MODE_PRX;
	config.selective_auto_ack = true;

	err = esb_init(&config);

	if (err) {
		return err;
	}

	err = esb_set_base_address_0(base_addr_0);
	if (err) {
		return err;
	}

	err = esb_set_base_address_1(base_addr_1);
	if (err) {
		return err;
	}

	err = esb_set_prefixes(addr_prefix, 1);
	if (err) {
		return err;
	}

	err = esb_enable_pipes(1<< NETWORK_PIPE);
	if (err) {
		return err;
	}

	return 0;
}

void main(void)
{
	k_msleep(1000);

	printk("Hello world from CPUNET %s\n", CONFIG_BOARD);
	int err;
	LOG_INF("Enhanced ShockBurst ptx sample");

	gpio_pin_configure_dt(&led0, GPIO_OUTPUT_ACTIVE);
	LOG_INF("Led configured");
	
	ret = ipc_service_open_instance(ipc0_instance);
	if ((ret < 0) && (ret != -EALREADY)) {
		LOG_INF("ipc_service_open_instance() failure - ret: %d", ret);
		return;
	}
	LOG_INF("IPC service opened complete - ret: %d", ret);

	ret = ipc_service_register_endpoint(ipc0_instance, &ep, &ep_cfg);
	if (ret < 0) {
		LOG_ERR("ipc_service_register_endpoint() failure");
		return;
	}
	LOG_INF("EP Registered complete");


	err = clocks_start();
	if (err) {
		return;
	}
	LOG_INF("Clock started");

	err = esb_initialize();
	if (err) {
		LOG_ERR("ESB initialization failed, err %d", err);
		return;
	}
	LOG_INF("Initialization complete");

	esb_start_rx();

	tx_payload.length = MAX_PAYLOAD_LEN;
    tx_payload.pipe = NETWORK_PIPE;
    tx_payload.noack = true;

	k_timer_init(&time_sync_timer, time_sync_expiry, NULL);
	k_timer_start(&time_sync_timer, K_USEC(TIME_SYNC), K_USEC(TIME_SYNC));

	k_sem_take(&bound_sem, K_FOREVER);
	
	while (1) {
		
		uint8_t a = 1;

		printk("Sending message\n");
		ret = ipc_service_send(&ep, &a, MAX_PAYLOAD_LEN);
		printk("IPC SEND RET MSG: %d\n",ret);

		k_msleep(100);

		// This should send 10 messages per second to the AppCore

	}
}
